/*
 * dt_cbn.h
 *
 *  Created on: Jan 9, 2012
 *      Author: Thomas Sakoparnig
 *      based on the ct_cbn package
 */

#ifndef DT_CBN_H_
#define DT_CBN_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <omp.h>

#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_multimin.h>

#include "queue.h"

#define DOUBLE_FORMAT "%.10g"

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )
#define MIN(a,b) ( (a) < (b) ? (a) : (b) )

int verbose;

gsl_rng *RNG;  // random number generator

int** GENOTYPE;  // GENOTYPE[i] is the integer i in binary (as int array)


typedef struct {
	int** P;  // cover relations of the event poset
	int n;  // number of events
	int* lin_ext;  // linear extension of the poset

	int* J_P;  // lattice of order ideals (genotypes)
	int m;  // lattice size

	double eps;  // error tolerance used for constructing P

	int* N_pa;  // number of parents for each genotype
	int** pa;  // list of parents
	int** pa_diff;  // index in which parent and child differ

	int* N_ch; // number of children in lattice
	int** ch_diff;  // index in which parent and child differ
} model;


typedef struct {
	int *g;  // genotype
	int is_compatible;  // compatible with model?
	int** Q;  // induced refinement of poset
	int* J_Q;  // corresponding sublattice
	int count;  // number of observations of this type
} data;

typedef struct {
	model* M; //model
	data* D; //data
	int* N_u; //No obs
} gsl_parameters;


int* get_int_array(const int n)
{
	int* x = calloc(n, sizeof(int));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	return x;
}


unsigned int* get_uint_array(const int n)
{
	unsigned int* x = calloc(n, sizeof(unsigned int));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	return x;
}


double* get_double_array(const int n)
{
	double* x = calloc(n, sizeof(double));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	return x;
}


int** get_int_matrix(const int m, const int n)
{
	int** x = malloc(m * sizeof(int *));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	int i;
	for (i=0; i<m; i++)
	{
		x[i] = calloc(n, sizeof(int));
		if (x[i] == NULL)
		{
			fprintf(stderr, "Error: Out of memory!\n");
			exit(1);
		}
	}

	return x;
}


double** get_double_matrix(const int m, const int n)
{
	double** x = malloc(m * sizeof(double *));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	int i;
	for (i=0; i<m; i++)
	{
		x[i] = calloc(n, sizeof(double));
		if (x[i]  == NULL)
		{
			fprintf(stderr, "Error: Out of memory!\n");
			exit(1);
		}
	}

	return x;
}

double*** get_double_cube(const int m, const int n, const int l)
{
	double*** x = malloc(m * sizeof(double **));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	int i;
	for (i=0; i<m; i++)
	{
		x[i] = get_double_matrix(n,l);
		if (x[i]  == NULL)
		{
			fprintf(stderr, "Error: Out of memory!\n");
			exit(1);
		}
	}

	return x;
}

int*** get_int_cube(const int m, const int n, const int l)
{
	int*** x = malloc(m * sizeof(int **));

	if (x == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	int i;
	for (i=0; i<m; i++)
	{
		x[i] = get_int_matrix(n,l);
		if (x[i]  == NULL)
		{
			fprintf(stderr, "Error: Out of memory!\n");
			exit(1);
		}
	}

	return x;
}

void print_int_array(int* x, int n)
{
	int j;

	if (n > 0)
	{
		for (j=0; j<n-1; j++)
			printf("%d ", x[j]);
		printf("%d", x[n-1]);
	}
	printf("\n");
}



void print_int_matrix(int** X, int m, int n)
{
	int i;

	for (i=0; i<m; i++)
		print_int_array(X[i], n);

}


void print_double_array(double* x, int n)
{
	int j;

	if (n > 0)
	{
		for (j=0; j<n; j++)
		{
			printf(DOUBLE_FORMAT, x[j]);
			if (j < n-1)
				printf("\t");
		}
	}
	printf("\n");
}



void print_double_matrix(double** X, int m, int n)
{
	int i;

	for (i=0; i<m; i++)
		print_double_array(X[i], n);

}



void write_poset(int k, char* filestem, int** P, int n, int b)
{
	int i, j;

	char filename[255];

	if (b >= 0)
		sprintf(filename, "%s/b%09d.poset", filestem, b);
	else
		sprintf(filename, "%s/%09d.poset", filestem, k);

	FILE *output;
	if ( (output = fopen(filename, "w")) == NULL )
	{
		fprintf(stderr, "Error:  Could not write to file %s\n", filename);
		fprintf(stderr, "        Make sure the directory '%s' exists.\n", filestem);
		exit(1);
	}

	fprintf(output, "%d\n", n);
	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
			if (P[i][j])
				fprintf(output, "%d %d\n", i+1, j+1);

	fprintf(output, "0 0\n");
	fclose(output);
}



void write_patterns(char* filestem, int** pat, int N, int n)
{
	int i, j;

	char suffix[15] = ".sim.pat";
	//char suffix[15] = ".pat";
	char *filename = (char *) calloc(strlen(filestem) + strlen(suffix) + 1, sizeof(char));
	strcat(filename, filestem);
	strcat(filename, suffix);

	FILE *output;
	if ( (output = fopen(filename, "w")) == NULL )
	{
		fprintf(stderr, "Error:  Could not write to file %s\n", filename);
		exit(1);
	}

	fprintf(output, "%d %d\n", N, n+1);
	for (i=0; i<N; i++)
	{
		for (j=0; j<n; j++)
			fprintf(output, "%d ", pat[i][j]);
		fprintf(output, "%d\n", pat[i][n]);
	}

	fclose(output);
}




inline void genotype_of(int index, int* x, int n)
{
	int i;

	for (i=n-1; i>=0; i--)
	{
		x[i] = index % 2;
		index = index / 2;
	}

}



int pow2(int k)
{
	int i, pow = 1;

	for (i=0; i<k; i++)
		pow *= 2;

	return(pow);
}



void precompute_binary(const int n)
{
	int i, j;
	int m = pow2(n);
	int *g = get_int_array(n);

	GENOTYPE = get_int_matrix(m, n);

	for (i=0; i<m; i++)
	{
		genotype_of(i, g, n);
		for (j=0; j<n; j++)
			GENOTYPE[i][j] = g[j];
	}
	free(g);
}



int index_of(int* x, int n)
{
	int i, index = 0;

	for (i=0; i<n; i++)
	{
		if (x[i] == 1)
			index += pow2(n-1-i);
	}

	return index;
}

int is_equal_int_matrix(int** A, int** B, int n)
{
	int i, j;

	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
			if (A[i][j] != B[i][j])
				return 0;

	return 1;
}



void boolean_matrix_sum(int** A, int** B, int** C, int n)
{
	/*
      Boolean matrix sum  A + B = C
	 */

	int i, j;

	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
		{
			C[i][j] = A[i][j] + B[i][j];
			C[i][j] = C[i][j] ? 1 : 0;
		}

}

void boolean_matrix_product(int** A, int** B, int** C, int n)
{
	/*
      Boolean matrix product  A * B = C
	 */

	int i, j, k;

	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
		{
			C[i][j] = 0;
			for (k=0; k<n; k++)
				C[i][j] += A[i][k] * B[k][j];
			C[i][j] = C[i][j] ? 1 : 0;
		}

}

void free_int_matrix( int** A, int n) {
	int j;

	for (j=0; j<n; j++)
		free(A[j]);
	free(A);
}

void free_double_matrix( double** A, int n) {
	int j;

	for (j=0; j<n; j++)
		free(A[j]);
	free(A);
}

void free_int_cube( int*** A, int n, int m) {
	int j;

	for (j=0; j<n; j++) {
		//printf("before free int mat\n");
		free_int_matrix(A[j], m );
	}
	//printf("before free A\n");
	//free(A);
}



void free_poset(model* M)
{
	free_int_matrix( M->P, M->n );
	free(M->lin_ext);
}



void free_lattice(model* M)
{
	int i;

	free(M->J_P);
	for (i=0; i<M->m; i++)
	{
		free(M->pa[i]);
		free(M->pa_diff[i]);
		free(M->ch_diff[i]);
	}
	free(M->pa);
	free(M->pa_diff);
	free(M->N_pa);
	free(M->ch_diff);
}

void free_lattice_children(model* M)
{
	int i;

	free(M->J_P);
	free_int_matrix( M->ch_diff, M->m );
	free(M->N_ch);
}



void free_model(model* M)
{
	free_poset(M);
	free_lattice(M);
}



void print_model(model* M)
{
	int i;
	int m = M->m;
	int n = M->n;
	int* g = get_int_array(n);

	printf("\n\nMODEL\n\n");
	printf("\nP =\n");
	print_int_matrix(M->P, n, n);
	printf("\n");

	printf("lattice size, m = %d\nsorted lattice = \n", m);
	for (i=0; i<m; i++)
	{
		printf("%d\t", i);
		genotype_of(M->J_P[i], g, n);
		print_int_array(g, n);
	}
	//print_int_array(M->J_P, m);

	printf("\nlinear extension of the poset = ");
	print_int_array(M->lin_ext, n);

	for (i=0; i<m; i++)
	{
		printf("\nparents of %5d = ", i);
		//genotype_of(M->J_P[i], g, n+1);
		//print_int_array(g, n+1);
		print_int_array(M->pa[i], M->N_pa[i]);
		printf("differing events = ");
		print_int_array(M->pa_diff[i], M->N_pa[i]);
	}
	printf("\n\n");

	for (i=0; i<m; i++)
	{
		printf("differing events to children of %5d = ", i);
		//genotype_of(M->J_P[i], g, n+1);
		//print_int_array(g, n+1);
		print_int_array(M->ch_diff[i], M->N_ch[i]);
	}

	printf("\n");

	free(g);
}



void print_data(data* D, int N_u, int n, int m)
{
	int k;

	printf("\n\nDATA\n\n");
	for (k=0; k<N_u; k++)
	{
		printf("# %d\n", k);
		printf("g = ");  print_int_array(D[k].g, n);
		//printf("Q =\n");  print_int_matrix(D[k].Q, n, n);
		//printf("J_Q = ");  print_int_array(D[k].J_Q, m);
		printf("count = %d\n", D[k].count);
		printf("is compatible = %d\n", D[k].is_compatible);
		printf("--------------------\n");
	}

}



void free_data(data* D, int N_u, int n)
{
	int k, j;

	for (k=0; k<N_u; k++)
	{
		free(D[k].g);
		for (j=0; j<=n; j++)
			free(D[k].Q[j]);
		free(D[k].Q);
		free(D[k].J_Q);
	}
	free(D);
}



int** read_patterns(char* filestem, int* N, int n)
{
	int j, k, p;

	char suffix[15] = ".pat";
	char *filename = (char *) calloc(strlen(filestem) + strlen(suffix) + 1, sizeof(char));
	strcat(filename, filestem);
	strcat(filename, suffix);

	FILE *input;
	if ( (input = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "Error:  Could not read %s\n", filename);
		exit(1);
	}

	/* Read dimensions */
	fscanf(input, "%d %d", N, &p);
	if (verbose) printf("\nreading data from file %s :  %d samples, %d events ...\n\n", filename, *N, p-1);
	if (*N < 1)
	{
		fprintf(stderr, "Error:  Less than one data point!\n");
		exit(1);
	}
	if (n != p)
	{
		fprintf(stderr, "Error:  Number of events in poset and data do not match!\n");
		exit(1);
	}

	int** pat = get_int_matrix(*N, p);

	/* Read patterns */
	int x;
	for (k=0; k<*N; k++)
	{
		for (j=0; j<n; j++)
		{
			if (fscanf(input,"%d ", &x) == 1)
			{
				if ((x != 0) && (x != 1))
					x = -1;
				pat[k][j] = x;
			}
			else
			{
				fprintf(stderr, "Error reading data from %s!\n", filename);
				exit(1);
			}

		}
	}

	return pat;
}

void read_poset(char* filestem, model* M)
{
	int left, right;

	char suffix[15] = ".poset";
	char *filename = (char *) calloc(strlen(filestem) + strlen(suffix) + 1, sizeof(char));
	strcat(filename, filestem);
	strcat(filename, suffix);

	FILE *input;
	if ( (input = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "Error:  Could not read %s\n", filename);
		exit(1);
	}

	/* Read number of relations */
	int n;
	fscanf(input, "%d", &n);
	if (verbose)  printf("n = %d events\n\n", n);
	if ((n < 1) || (n > 25))
	{
		fprintf(stderr, "Error:  Number of events is %d.  Supported range is {1, ..., 14}.\n", n);
		exit(1);
	}

	M->n = n;
	M->P = get_int_matrix(n+1, n+1);

	/* Read partial orderings from file */
	fscanf(input,"%d %d", &left, &right);
	while (left != 0)
	{
		if (verbose)  printf("%d --> %d\n", left, right);  // i.e., left < right
		if ((left > n) || (right > n) || (left < 0) || (right < 1))
		{
			fprintf(stderr, "Error:  Undefined event in %s!\n", filename);
			exit(1);
		}
		M->P[left][right] = 1;
		fscanf(input,"%d %d", &left, &right);
	}

	fclose(input);
}

void print_genotype(int* x, int n)
{
	int i;

	for (i=0; i<n; i++)
		printf("%d", x[i]);

}



int* bfs_order_ideals(int** poset, const int len, int* count, int* lin_ext)
{
	// implements BFS in the genotype lattice

	int i, j, k, is_compatible;
	int lin_ext_size = 0;

	queue q;
	int g_idx, new_idx;  // current, new genotype index
	int* g = get_int_array(len);  // current genotype

	int* lattice = malloc(sizeof(int));

	int* added = get_int_array(pow2(len));  // records added genotypes
	/* TODO: make this dynamic, e.g., using a list */

	init_queue(&q);
	new_idx = 0;  // wild type 0...0
	enqueue(&q, new_idx);
	added[new_idx] = 1;

	*count = 0;
	while (empty(&q) == FALSE) {
		g_idx = dequeue(&q);
		genotype_of(g_idx, g, len);

		// visit...
		(*count)++;
		if ((lattice = (int*) realloc(lattice, (*count) * sizeof(int))) == NULL)
		{
			fprintf(stderr, "Error: Out of memory!\n");
			exit(1);
		}
		lattice[(*count)-1] = g_idx;

		// linear extension
		for (i=1; i<len; i++)  // exclude 0
		{
			if (g[i] == 1)
			{
				int is_in = 0;
				for (j=0; j<lin_ext_size; j++)
					if (lin_ext[j] == i)
					{
						is_in = 1;
						break;
					}
				if (! is_in)  // add to linear extension:
				{
					lin_ext[lin_ext_size] = i;
					lin_ext_size++;
				}
			}
		}


		// generate children:
		for (i=0; i<len; i++)
		{
			if (g[i] == 0)
			{
				g[i] = 1;  // try this new event
				new_idx = index_of(g, len);

				// check compatibility:
				is_compatible = 1;
				for (k=0; k<len; k++)
				{
					if ((poset[k][i] == 1) && (g[k] == 0))
					{
						is_compatible = 0;
						break;
					}
				}

				if ((is_compatible) && (! added[new_idx]))
					/* add if compatible and really new */
				{
					enqueue(&q, new_idx);
					added[new_idx] = 1;
				}

				g[i] = 0;  // undo event
			}
		}
	}

	free(g);
	free(added);

	return(lattice);
}



int norm1(const int g_idx, const int n)
{
	int i;
	int *g = get_int_array(n);

	genotype_of(g_idx, g, n);

	int norm1 = 0;
	for(i=0; i<n; i++)
		norm1 += g[i];

	free(g);

	return norm1;
}


double power(double m, int n)
{
	int i;
	double x=1;
	for(i=0;i<n;i++)
		x*=m;
	return x;
}

unsigned hamdist(unsigned x, unsigned y)
{
	unsigned dist = 0, val = x ^ y;

	while(val)
	{
		++dist;
		val &= val - 1;
	}

	return dist;
}

int hamming_distance(int g_idx, int h_idx, int* diff_idx, int n)
{
	int i;
	int *g = get_int_array(n);
	int *h = get_int_array(n);

	genotype_of(g_idx, g, n);
	genotype_of(h_idx, h, n);

	int dist = 0;
	for(i=0; i<n; i++)
		if (g[i] != h[i])
		{
			dist++;
			*diff_idx = i;
		}

	free(g);
	free(h);

	return dist;
}

void parents_dt(model* M)
{
	int i, j, k, c;
	int m = M->m;
	int n = M->n;

	M->N_pa = get_int_array(m);
	M->pa = malloc(m * sizeof(int*));
	M->pa_diff = malloc(m * sizeof(int*));
	if ((M->pa == NULL) || (M->pa_diff == NULL))
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	// count parents:
	for (i=0; i<m; i++)
	{
		M->N_pa[i] = 0;  // number of parents
		k = i-1;
		while ((k >=0) && (norm1(M->J_P[k], n) >= (norm1(M->J_P[i], n) - 1)))
		{
			//if (hamming_distance(M->J_P[i], M->J_P[k], &j, n+1) == 1)
			if (hamdist(M->J_P[i], M->J_P[k]) == 1)
			{
				M->N_pa[i]++;  // found a parent!
			}
			k--;
		}
	}

	// list parents:
	for (i=0; i<m; i++)
	{
		M->pa[i] = get_int_array(M->N_pa[i]);
		M->pa_diff[i] = get_int_array(M->N_pa[i]);
		// generate parents in sublattice
		k = i-1;
		c = 0;
		while ((k >=0) && (norm1(M->J_P[k], n) >= (norm1(M->J_P[i], n) - 1)))
		{
			if(hamming_distance(M->J_P[i], M->J_P[k], &j, n) == 1)
			{
				M->pa[i][c] = k;  // save index of parent in J_P[]
				M->pa_diff[i][c] = j;  // save index by which parent and child differ
				c++;
			}
			k--;
		}
	}

}



void children_dt(model* M)
{
	int i, j, k, c;
	int m = M->m;
	int n = M->n;

	M->N_ch = get_int_array(m);
	M->ch_diff = malloc(m * sizeof(int*));
	if (M->ch_diff == NULL)
	{
		fprintf(stderr, "Error: Out of memory!\n");
		exit(1);
	}

	// count children:
	for (i=0; i<m; i++)
	{
		M->N_ch[i] = 0;
		/* generate children in lattice */
		k = i + 1;
		while ( (k < m) && (norm1(M->J_P[k], n) <= (norm1(M->J_P[i], n) + 1)) )
		{
			//if (hamming_distance(M->J_P[i], M->J_P[k], &j, n+1) == 1)
			if (hamdist(M->J_P[i], M->J_P[k]) == 1)
			{
				M->N_ch[i]++;
			}
			k++;
		}
	}

	// list children:
	for (i=0; i<m; i++)
	{
		M->ch_diff[i] = get_int_array(M->N_ch[i]);
		/* generate children in lattice */
		k = i + 1;
		c = 0;
		while ( (k < m) && (norm1(M->J_P[k], n) <= (norm1(M->J_P[i], n) + 1)) )
		{
			if (hamming_distance(M->J_P[i], M->J_P[k], &j, n) == 1)
			{
				M->ch_diff[i][c] = j;
				c++;
			}
			k++;
		}
	}

}

void read_poset_dt(char* filestem, model* M)
{
	int left, right;

	char suffix[15] = ".poset";
	char *filename = (char *) calloc(strlen(filestem) + strlen(suffix) + 1, sizeof(char));
	strcat(filename, filestem);
	strcat(filename, suffix);

	FILE *input;
	if ( (input = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "Error:  Could not read %s\n", filename);
		exit(1);
	}

	/* Read number of relations */
	int n;
	fscanf(input, "%d", &n);
	if (verbose)  printf("n = %d events\n\n", n);
	if ((n < 1) || (n > 25))
	{
		fprintf(stderr, "Error:  Number of events is %d.  Supported range is {1, ..., 14}.\n", n);
		exit(1);
	}

	M->n = n;
	M->P = get_int_matrix(n, n);

	/* Read partial orderings from file */
	fscanf(input,"%d %d", &left, &right);
	while (left != 0)
	{
		if (verbose)  printf("%d --> %d\n", left, right);  // i.e., left < right
		if ((left > n) || (right > n) || (left < 0) || (right < 1))
		{
			fprintf(stderr, "Error:  Undefined event in %s!\n", filename);
			exit(1);
		}
		M->P[left-1][right-1] = 1;
		fscanf(input,"%d %d", &left, &right);
	}

	fclose(input);
}

data* make_data_set(int** pat, int N, int n, int* N_u, int* pat_idx)
{
	int j, k, l, skip;

	int* idx = get_int_array(N);
	int* count = get_int_array(N);

	for (k=0; k<N; k++)
		idx[k] = index_of(pat[k], n);

	// count unique patterns:
	int c = 0;
	for (k=0; k<N; k++)
	{
		skip = 0;
		for (l=0; l<k; l++)
		{
			if (idx[k] == idx[l]) // observed before?
			{
				count[l]++;
				pat_idx[k] = pat_idx[l];
				skip = 1;
				break;
			}
		}
		if (! skip) // Unique!
		{
			count[k]++;
			pat_idx[k] = c;
			c++;
		}
	}

	*N_u = 0;
	for (k=0; k<N; k++)
		*N_u += (count[k] > 0);

	if (verbose)  printf("N_u = %d unique patterns\n", *N_u);

	data* D = calloc(*N_u, sizeof(data));

	c = 0;
	for (k=0; k<N; k++)
	{
		//    pat_idx[k] = c;
		if (count[k] > 0)
		{
			//	  pat_idx[k]=c;
			D[c].count = count[k];
			D[c].g = get_int_array(n);
			for (j=0; j<n; j++)
				D[c].g[j] = pat[k][j];
			c++;
		}
	}

	return D;
}

void compatibility(data* D, int N_u, model* M)
{
	int i, j, k;

	for (k=0; k<N_u; k++)
	{
		D[k].is_compatible = 1;
		i = 0;
		while ((i < M->n) && D[k].is_compatible)
		{
			j = 0;
			while ((j < M->n) && D[k].is_compatible)
			{
				if (M->P[i][j] && (! D[k].g[i]) && D[k].g[j])
					/* It is sufficient to check the cover relations! */
					//if (lt_poset(i, j, M->P, M->n) && (! D[k].g[i]) && D[k].g[j])
					D[k].is_compatible = 0;
				j++;
			}
			i++;
		}
	}

}


/**
 * Pr(X|M,theta) for all X in J(P)
 * @param theta IN mutation probabilities; size n
 * @param M IN the model
 * @param Prob OUT the probabilities of all valid (according to the poset) genotypes; size m
 * @param theta_exit IN the exit probabilities for all elements in the genotype lattice; size: m
 */
void compute_all_cbn_prob( double* theta, model* M, double* Prob, double* theta_exit ) {

	int m = M->m;
	int n = M->n;

	int i,j,c,k;

	Prob[0] = 1.0;

	for (i=0; i < m; i++)
	{
		//printf("%d ",M->J_P[i]);
		//print_int_array(GENOTYPE[M->J_P[i]],n);
		double t = 1.0;
		for (c=0; c<n; c++)
		{
			if( GENOTYPE[M->J_P[i]][c] ) {
				t *= theta[c];
			}
		}
		Prob[i] = t * theta_exit[i];

	}

	double p = 0;
	for(i=0;i<m;i++) {
		p += Prob[i];
	}
	//printf(DOUBLE_FORMAT ,p);
}

/*
 * IN double *theta : mutation probabilities; size: n
 * IN model *M : the model
 * OUT double * theta_exit: the exit probabilities for all elements in the genotype lattice; size: m
 */
void compute_theta_exit(double *theta, model* M, double* theta_exit)
{
	int i, j, c;

	for (i=0; i<M->m; i++)
	{
		theta_exit[i] = 1.0;

		for (c=0; c<M->N_ch[i]; c++)
		{
			j = M->ch_diff[i][c];  // index of differing event
			theta_exit[i] *= (1-theta[j]);

		}
	}
}

// Compute the conditional probability Prob[Z|X,epsilon]
void compute_cond_error_prob(model* M, data* D, int N_u, double epsilon, double** condprob)
{
	int n = M->n;
	int m = M->m;

	int i,k,d,index;


	for (k=0; k<N_u; k++)   // All patients
	{
		//prob_tmp = 0;
		index = index_of(D[k].g, n);

		for (i=1; i<m; i++) // All patterns
		{
			condprob[i][k] = 0.0;
			d = hamdist(M->J_P[i], index);
			condprob[i][k] = power(epsilon, d) * power(1 - epsilon, n - d);
			//prob_tmp += condprob[i][k];
		}
		// Now divide by Prob[Y_k]
		//for (i=1; i<m; i++) // All patterns
		//condprob[i][k] /= prob_tmp;
	}
}


void clone_poset( model* M_src, model* M_dest ) {

	int i,j;
	int n = M_src->n;
	//int m = M_src->m;
	M_dest->n = n;
	M_dest->P = get_int_matrix(n, n);
	M_dest->lin_ext = get_int_array( n );
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			M_dest->P[i][j] = M_src->P[i][j];
		}
		M_dest->lin_ext[i] = M_src->lin_ext[i];
	}
	/*
	M_dest->m = m;
	M_dest->J_P = get_int_array( m );
	for( i=0;i<m;i++ ) {
		M_dest->J_P[i] = M_src->J_P[i];
	}
	M_dest->eps = M_src->eps;
	 */

}

void transitive_closure(int** A, int**T, int n)
{
	/*
      T is the transitive closure of the relation A
	 */

	int** R = get_int_matrix(n, n);
	int** S = get_int_matrix(n, n);

	int i, j, k = 0;

	// initialize T = A
	for (i=0; i<n; i++)
		for (j=0; j<n; j++)
			T[i][j] = A[i][j];

	while ((k == 0) || (! is_equal_int_matrix(S, T, n)))
	{
		// S = T
		for (i=0; i<n; i++)
			for (j=0; j<n; j++)
				S[i][j] = T[i][j];

		// T = A*S + A
		boolean_matrix_product(A, S, R, n);
		boolean_matrix_sum(R, A, T, n);

		k++;
	}

	free_int_matrix(R, n);
	free_int_matrix(S, n);

}

/**
 *
 * @param P transitive closure to cover relations
 * @param n
 * @param C changes
 * @return status
 */
int reduce_to_cover_relations ( int** P, int n, int** C )
{
	int i,j,k;
	queue q;
	int stat = 0;

	// Sample all nodes
	for (i=0; i<n; i++)
	{
		int* visit = get_int_array(n);
		init_queue(&q);

		// Fill queue with children of i
		for (j=0; j<n; j++)
			if (P[i][j])
			{
				enqueue(&q,j);
			}

		// Walk through grandchildren
		while (empty(&q) == FALSE)
		{
			j = dequeue(&q);
			for (k=0; k<n; k++)
			{
				if (P[j][k] && !(visit[k]))
				{
					visit[k] = 1;
					enqueue(&q,k);

					// Remove non-cover relations
					if (P[i][k])
					{
						P[i][k] = 0;
						C[i][k] = 1;
						stat = 1; // Report changes
					}

					// Check if cyclic
					if (P[k][i]) {
						free(visit);
						return 2; // Fatal
					}
				}
			}
		}
		free(visit);
	}
	return stat;
}

#endif /* DT_CBN_H_ */
